'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _fs;

function _load_fs() {
  return _fs = _interopRequireDefault(require('fs'));
}

var _path;

function _load_path() {
  return _path = _interopRequireDefault(require('path'));
}

var _glob;

function _load_glob() {
  return _glob = _interopRequireDefault(require('glob'));
}

var _object;

function _load_object() {
  return _object = _interopRequireDefault(require('./object'));
}

var _ref;

function _load_ref() {
  return _ref = _interopRequireDefault(require('./ref'));
}

var _constants;

function _load_constants() {
  return _constants = require('../../constants');
}

var _exceptions;

function _load_exceptions() {
  return _exceptions = require('../exceptions');
}

var _utils;

function _load_utils() {
  return _utils = require('../../utils');
}

var _scope;

function _load_scope() {
  return _scope = require('../../scope');
}

var _component;

function _load_component() {
  return _component = _interopRequireDefault(require('../models/component'));
}

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

class Repository {

  constructor(scope) {
    let objectTypes = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];
    this.objects = [];
    this._cache = {};

    this.scope = scope;
    this.types = objectTypes.reduce((map, objectType) => {
      map[objectType.name] = objectType;
      return map;
    }, {});
  }

  ensureDir() {
    return (0, (_utils || _load_utils()).mkdirp)(this.getPath());
  }

  getPath() {
    return (_path || _load_path()).default.join(this.scope.getPath(), (_constants || _load_constants()).OBJECTS_DIR);
  }

  objectPath(ref) {
    const hash = ref.toString();
    return (_path || _load_path()).default.join(this.getPath(), hash.slice(0, 2), hash.slice(2));
  }

  load(ref) {
    if (this.getCache(ref)) return Promise.resolve(this.getCache(ref));
    return (0, (_utils || _load_utils()).readFile)(this.objectPath(ref)).then(fileContents => {
      return (_object || _load_object()).default.parseObject(fileContents, this.types);
    }).catch(() => null);
  }

  list() {
    // @TODO - write
    const filterComponents = refs => refs.filter(ref => ref instanceof (_component || _load_component()).default);

    return new Promise((resolve, reject) => {
      return (0, (_glob || _load_glob()).default)((_path || _load_path()).default.join('*', '*'), { cwd: this.getPath() }, (err, matches) => {
        if (err) reject(err);
        const refs = matches.map(str => str.replace((_path || _load_path()).default.sep, ''));
        Promise.all(refs.map(ref => this.load(ref))).then(filterComponents).then(resolve);
      });
    });
  }

  remove(ref) {
    return (0, (_utils || _load_utils()).removeFile)(this.objectPath(ref), true);
  }

  removeMany(refs) {
    return Promise.all(refs.map(ref => this.remove(ref)));
  }

  loadRaw(ref) {
    return (0, (_utils || _load_utils()).readFile)(this.objectPath(ref));
  }

  loadSync(ref) {
    try {
      return (_object || _load_object()).default.parseSync((_fs || _load_fs()).default.readFileSync(this.objectPath(ref)), this.types);
    } catch (err) {
      throw new (_exceptions || _load_exceptions()).HashNotFound(ref.toString());
    }
  }

  setCache(object) {
    this._cache[object.hash().toString()] = object;
    return this;
  }

  getCache(ref) {
    return this._cache[ref.toString()];
  }

  add(object) {
    if (!object) return this;
    this.objects.push(object);
    this.setCache(object);
    return this;
  }

  /**
   * alias to `load`
   */
  findOne(ref) {
    return this.load(ref);
  }

  findMany(refs) {
    return Promise.all(refs.map(ref => this.load(ref)));
  }

  persist() {
    // @TODO handle failures
    return Promise.all(this.objects.map(object => this.persistOne(object)));
  }

  persistOne(object) {
    return object.compress().then(contents => {
      const options = {};
      if (this.scope.groupName) options.gid = (0, (_utils || _load_utils()).resolveGroupId)(this.scope.groupName);
      return (0, (_utils || _load_utils()).writeFile)(this.objectPath(object.hash()), contents, options);
    });
  }
}
exports.default = Repository;