'use strict';

var _fs;

function _load_fs() {
  return _fs = _interopRequireDefault(require('fs'));
}

var _path;

function _load_path() {
  return _path = _interopRequireDefault(require('path'));
}

var _chai;

function _load_chai() {
  return _chai = require('chai');
}

var _jsdoc;

function _load_jsdoc() {
  return _jsdoc = require('../../src/jsdoc');
}

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

describe('JSDoc Parser', () => {
  describe('parse()', () => {
    describe('Function Declaration', () => {
      let doclet;
      before(function () {
        const functionDeclarationFile = (_path || _load_path()).default.join(__dirname, 'fixtures', 'functionDeclaration.js');
        const functionDeclaration = (_fs || _load_fs()).default.readFileSync(functionDeclarationFile).toString();
        const doclets = (0, (_jsdoc || _load_jsdoc()).parser)(functionDeclaration);
        (0, (_chai || _load_chai()).expect)(doclets).to.be.a('array');
        (0, (_chai || _load_chai()).expect)(doclets).to.have.lengthOf(1);
        doclet = doclets[0];
      });
      it('should have all the proper keys', () => {
        (0, (_chai || _load_chai()).expect)(doclet).to.include.keys('name', 'description', 'returns', 'args', 'access', 'examples', 'static');
      });
      it('should extract the correct name', () => {
        (0, (_chai || _load_chai()).expect)(doclet.name).to.equal('add');
      });
      it('should extract the correct description', () => {
        (0, (_chai || _load_chai()).expect)(doclet.description).to.equal('Adds two numbers.');
      });
      it('should extract the correct args', () => {
        const args = doclet.args;
        (0, (_chai || _load_chai()).expect)(args).to.be.a('array');
        (0, (_chai || _load_chai()).expect)(args).to.have.lengthOf(2);
        for (const arg of args) {
          (0, (_chai || _load_chai()).expect)(arg).to.include.keys('name', 'type', 'description');
        }
        (0, (_chai || _load_chai()).expect)(args[0].name).to.equal('a');
        (0, (_chai || _load_chai()).expect)(args[1].name).to.equal('b');
        (0, (_chai || _load_chai()).expect)(args[0].description).to.equal('The first number in an addition.');
        (0, (_chai || _load_chai()).expect)(args[1].description).to.equal('The second number in an addition.');
        (0, (_chai || _load_chai()).expect)(args[0].type).to.equal('number');
        (0, (_chai || _load_chai()).expect)(args[1].type).to.equal('number');
      });
      it('should extract the correct returns', () => {
        (0, (_chai || _load_chai()).expect)(doclet.returns).to.include.keys('type', 'description');
        (0, (_chai || _load_chai()).expect)(doclet.returns.description).to.equal('Returns the total.');
        (0, (_chai || _load_chai()).expect)(doclet.returns.type).to.equal('number');
      });
      it('should extract the correct access from the @public annotation', () => {
        (0, (_chai || _load_chai()).expect)(doclet.access).to.equal('public');
      });
      it('should extract the static attribute from the @static annotation', () => {
        (0, (_chai || _load_chai()).expect)(doclet.static).to.be.true;
      });
    });

    describe('Invalid code', () => {
      it('should returns an empty array', () => {
        const doclets = (0, (_jsdoc || _load_jsdoc()).parser)('this is an invalid code');
        (0, (_chai || _load_chai()).expect)(doclets).to.be.a('array');
        (0, (_chai || _load_chai()).expect)(doclets).to.have.lengthOf(0);
      });
    });

    describe('Method Declaration and Class Declaration', () => {
      let doclets;
      before(function () {
        const methodDeclarationFile = (_path || _load_path()).default.join(__dirname, 'fixtures', 'methodDeclaration.js');
        const methodDeclaration = (_fs || _load_fs()).default.readFileSync(methodDeclarationFile).toString();
        doclets = (0, (_jsdoc || _load_jsdoc()).parser)(methodDeclaration);
      });
      it('should be a populated array', () => {
        (0, (_chai || _load_chai()).expect)(doclets).to.be.an('array');
        (0, (_chai || _load_chai()).expect)(doclets).to.have.length.of.at.least(2);
      });
      xit('should recognize the Class Declaration first', () => {
        const doclet = doclets[0];
        (0, (_chai || _load_chai()).expect)(doclet).to.have.all.keys('name', 'description');
        (0, (_chai || _load_chai()).expect)(doclet.name).to.equal('Point');
        (0, (_chai || _load_chai()).expect)(doclet.description).to.equal('Class representing a point.');
      });
      it('should recognize the constructor', () => {
        const doclet = doclets[0];
        (0, (_chai || _load_chai()).expect)(doclet).to.have.property('name').that.equals('constructor');
        (0, (_chai || _load_chai()).expect)(doclet).to.have.property('description').that.equals('Create a point.');
        (0, (_chai || _load_chai()).expect)(doclet).to.have.property('access').that.equals('public');
        (0, (_chai || _load_chai()).expect)(doclet).to.have.property('static').that.is.false;
        (0, (_chai || _load_chai()).expect)(doclet).to.have.property('returns').that.is.an('object').that.is.empty;
        (0, (_chai || _load_chai()).expect)(doclet).to.have.property('examples').that.is.an('array').that.is.empty;
      });
      it('should extract the constructor\'s args correctly', () => {
        const doclet = doclets[0];
        (0, (_chai || _load_chai()).expect)(doclet).to.have.property('args').that.is.an('array').with.lengthOf(2);
        const args = doclet.args;
        for (const arg of args) {
          (0, (_chai || _load_chai()).expect)(arg).to.include.keys('name', 'type', 'description');
          (0, (_chai || _load_chai()).expect)(arg.type).to.equal('number');
        }
        (0, (_chai || _load_chai()).expect)(args[0].name).to.equal('x');
        (0, (_chai || _load_chai()).expect)(args[1].name).to.equal('y');
        (0, (_chai || _load_chai()).expect)(args[0].description).to.equal('The x value.');
        (0, (_chai || _load_chai()).expect)(args[1].description).to.equal('The y value.');
      });
      it('should recognize the getX method', () => {
        const doclet = doclets[1];
        (0, (_chai || _load_chai()).expect)(doclet).to.have.property('name').that.equals('getX');
        (0, (_chai || _load_chai()).expect)(doclet).to.have.property('description').that.equals('Get the x value.');
        (0, (_chai || _load_chai()).expect)(doclet).to.have.property('access').that.equals('public');
        (0, (_chai || _load_chai()).expect)(doclet).to.have.property('static').that.is.false;
        (0, (_chai || _load_chai()).expect)(doclet).to.have.property('returns').that.is.an('object').that.is.not.empty;
        (0, (_chai || _load_chai()).expect)(doclet).to.have.property('examples').that.is.an('array').that.is.empty;
        (0, (_chai || _load_chai()).expect)(doclet).to.have.property('args').that.is.an('array').that.is.empty;
      });
      it('should recognize the getY method', () => {
        const doclet = doclets[2];
        (0, (_chai || _load_chai()).expect)(doclet).to.have.property('name').that.equals('getY');
        (0, (_chai || _load_chai()).expect)(doclet).to.have.property('description').that.equals('Get the y value.');
      });
      it('should recognize the fromString method as the last doclet', () => {
        const doclet = doclets[doclets.length - 1];
        (0, (_chai || _load_chai()).expect)(doclet).to.have.property('name').that.equals('fromString');
        (0, (_chai || _load_chai()).expect)(doclet).to.have.property('description').that.equals('Convert a string containing two comma-separated numbers into a point.');
      });
    });

    describe('Variable Declaration', () => {
      let doclets;
      before(function () {
        const variableDeclarationFile = (_path || _load_path()).default.join(__dirname, 'fixtures', 'variableDeclaration.js');
        const variableDeclaration = (_fs || _load_fs()).default.readFileSync(variableDeclarationFile).toString();
        doclets = (0, (_jsdoc || _load_jsdoc()).parser)(variableDeclaration);
      });
      it('should be an array of one Doclet', () => {
        (0, (_chai || _load_chai()).expect)(doclets).to.be.an('array').and.to.have.lengthOf(1);
      });
      it('should parse the doc correctly', () => {
        const doclet = doclets[0];
        (0, (_chai || _load_chai()).expect)(doclet).to.have.property('name').that.equals('add');
        (0, (_chai || _load_chai()).expect)(doclet).to.have.property('description').that.equals('Adds two numbers.');
        (0, (_chai || _load_chai()).expect)(doclet).to.have.property('access').that.equals('public');
        (0, (_chai || _load_chai()).expect)(doclet).to.have.property('static').that.is.false;
        (0, (_chai || _load_chai()).expect)(doclet).to.have.property('returns').that.is.an('object').that.is.not.empty;
        (0, (_chai || _load_chai()).expect)(doclet).to.have.property('examples').that.is.an('array').that.is.not.empty;
        (0, (_chai || _load_chai()).expect)(doclet).to.have.property('args').that.is.an('array').that.have.lengthOf(2);
      });
    });

    describe('Various Param Types', () => {
      let args;
      before(function () {
        const file = (_path || _load_path()).default.join(__dirname, 'fixtures', 'variousParamTypes.js');
        const doclets = (0, (_jsdoc || _load_jsdoc()).parser)((_fs || _load_fs()).default.readFileSync(file).toString());
        (0, (_chai || _load_chai()).expect)(doclets).to.be.an('array').and.to.have.lengthOf(1);
        const doclet = doclets[0];
        (0, (_chai || _load_chai()).expect)(doclet).to.have.property('args').that.is.an('array').that.is.not.empty;
        args = doclet.args;
      });
      it('should recognize "*" as "*"', () => {
        const anyArg = args.find(arg => arg.name === 'anyType');
        (0, (_chai || _load_chai()).expect)(anyArg.type).to.equal('*');
      });
      it('should recognize "[]" as "[]"', () => {
        const anyArg = args.find(arg => arg.name === 'arrayType');
        (0, (_chai || _load_chai()).expect)(anyArg.type).to.equal('[]');
      });
      it('should recognize Union type correctly', () => {
        const anyArg = args.find(arg => arg.name === 'unionType');
        (0, (_chai || _load_chai()).expect)(anyArg.type).to.equal('(number|[])');
      });
      it('should recognize custom type correctly', () => {
        const anyArg = args.find(arg => arg.name === 'myCustomType');
        (0, (_chai || _load_chai()).expect)(anyArg.type).to.equal('CustomType');
      });
      it('should recognize Object type', () => {
        const anyArg = args.find(arg => arg.name === 'objectType');
        (0, (_chai || _load_chai()).expect)(anyArg.type).to.equal('Object');
      });
      it('should recognize Function type', () => {
        const anyArg = args.find(arg => arg.name === 'functionType');
        (0, (_chai || _load_chai()).expect)(anyArg.type).to.equal('Function');
      });
      it('should recognize Array of one type', () => {
        const anyArg = args.find(arg => arg.name === 'arrayOfType');
        (0, (_chai || _load_chai()).expect)(anyArg.type).to.equal('Array<string>');
      });
      it('should recognize Array of union', () => {
        const anyArg = args.find(arg => arg.name === 'arrayOfUnion');
        (0, (_chai || _load_chai()).expect)(anyArg.type).to.equal('Array<(number|Object)>');
      });
    });

    describe('Flow Type File', () => {
      it('should parse the file with no errors', () => {
        const file = (_path || _load_path()).default.join(__dirname, 'fixtures', 'flowTypeFile.js');
        const doclets = (0, (_jsdoc || _load_jsdoc()).parser)((_fs || _load_fs()).default.readFileSync(file).toString());
        (0, (_chai || _load_chai()).expect)(doclets).to.be.an('array').and.to.have.lengthOf(1);
        const doclet = doclets[0];

        (0, (_chai || _load_chai()).expect)(doclet).to.have.property('name').that.equals('first');
        (0, (_chai || _load_chai()).expect)(doclet).to.have.property('description').that.equals('returns the first element of an array reference.');
        (0, (_chai || _load_chai()).expect)(doclet).to.have.property('access').that.equals('public');
        (0, (_chai || _load_chai()).expect)(doclet).to.have.property('static').that.is.false;
        (0, (_chai || _load_chai()).expect)(doclet).to.have.property('returns').that.is.an('object').that.is.not.empty;
        (0, (_chai || _load_chai()).expect)(doclet).to.have.property('examples').that.is.an('array').that.is.not.empty;
        (0, (_chai || _load_chai()).expect)(doclet).to.have.property('args').that.is.an('array').that.is.not.empty;
      });
    });
  });
});
//# sourceMappingURL=parser.spec.js.map