'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _ramda;

function _load_ramda() {
  return _ramda = _interopRequireDefault(require('ramda'));
}

var _keyGetter;

function _load_keyGetter() {
  return _keyGetter = _interopRequireDefault(require('./key-getter'));
}

var _componentObjects;

function _load_componentObjects() {
  return _componentObjects = _interopRequireDefault(require('../../component-objects'));
}

var _exceptions;

function _load_exceptions() {
  return _exceptions = require('../exceptions');
}

var _bitId;

function _load_bitId() {
  return _bitId = require('../../../bit-id');
}

var _utils;

function _load_utils() {
  return _utils = require('../../../utils');
}

var _cliUtils;

function _load_cliUtils() {
  return _cliUtils = require('../../../cli/cli-utils');
}

var _component;

function _load_component() {
  return _component = _interopRequireDefault(require('../../../consumer/component'));
}

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const rejectNils = (_ramda || _load_ramda()).default.reject((_ramda || _load_ramda()).default.isNil);
const sequest = require('sequest');

function absolutePath(path) {
  if (!path.startsWith('/')) return `~/${path}`;
  return path;
}

function clean(str) {
  return str.replace('\n', '');
}

function errorHandler(err) {
  switch (err.code) {
    default:
      return new (_exceptions || _load_exceptions()).UnexpectedNetworkError();
    case 127:
      return new (_exceptions || _load_exceptions()).RemoteScopeNotFound();
    case 130:
      return new (_exceptions || _load_exceptions()).PermissionDenied();
  }
}

class SSH {

  constructor(_ref) {
    let path = _ref.path,
        username = _ref.username,
        port = _ref.port,
        host = _ref.host;

    this.path = path;
    this.username = username;
    this.port = port;
    this.host = host || '';
  }

  buildCmd(commandName) {
    for (var _len = arguments.length, args = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
      args[_key - 1] = arguments[_key];
    }

    function serialize() {
      return args.map(val => (0, (_utils || _load_utils()).toBase64)(val)).join(' ');
    }

    return `bit ${commandName} ${serialize()}`;
  }

  exec(commandName) {
    for (var _len2 = arguments.length, args = Array(_len2 > 1 ? _len2 - 1 : 0), _key2 = 1; _key2 < _len2; _key2++) {
      args[_key2 - 1] = arguments[_key2];
    }

    return new Promise((resolve, reject) => {
      const cmd = this.buildCmd(commandName, absolutePath(this.path || ''), ...args);
      this.connection(cmd, function (err, res, o) {
        if (!o) return reject(new (_exceptions || _load_exceptions()).UnexpectedNetworkError());
        if (err && o.code && o.code !== 0) return reject(errorHandler(err));
        return resolve(clean(res));
      });
    });
  }

  push(componentObjects) {
    return this.exec('_put', componentObjects.toString()).then(str => {
      try {
        return (_componentObjects || _load_componentObjects()).default.fromString((0, (_utils || _load_utils()).fromBase64)(str));
      } catch (err) {
        throw new (_exceptions || _load_exceptions()).NetworkError(str);
      }
    });
  }

  describeScope() {
    return this.exec('_scope').then(data => {
      return JSON.parse((0, (_utils || _load_utils()).fromBase64)(data));
    }).catch(() => {
      throw new (_exceptions || _load_exceptions()).RemoteScopeNotFound();
    });
  }

  list() {
    return this.exec('_list').then(str => {
      const components = (0, (_cliUtils || _load_cliUtils()).unpack)(str);
      return rejectNils(components.map(c => {
        return c ? (_component || _load_component()).default.fromString(c) : null;
      }));
    });
  }

  search(query, reindex) {
    return this.exec('_search', query, reindex.toString()).then(JSON.parse);
  }

  show(id) {
    return this.exec('_show', id.toString()).then(str => {
      const component = (0, (_cliUtils || _load_cliUtils()).unpack)(str);
      return str ? (_component || _load_component()).default.fromString(component) : null;
    });
  }

  fetch(ids) {
    let noDeps = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;

    let options = '';
    ids = ids.map(bitId => bitId.toString());
    if (noDeps) options = '-n';
    return this.exec(`_fetch ${options}`, ...ids).then(str => {
      const components = (0, (_cliUtils || _load_cliUtils()).unpack)(str);
      return components.map(raw => {
        return (_componentObjects || _load_componentObjects()).default.fromString(raw);
      });
    });
  }

  close() {
    this.connection.end();
    return this;
  }

  composeConnectionUrl() {
    return `${this.username}@${this.host}:${this.port}`;
  }

  connect(sshUrl, key) {
    this.connection = sequest.connect(this.composeConnectionUrl(), {
      privateKey: (0, (_keyGetter || _load_keyGetter()).default)(key)
    });

    return Promise.resolve(this);
  }
}
exports.default = SSH;
//# sourceMappingURL=ssh.js.map